# http.tcl
# Modified by: Nadim Parvez (UofC), May 5, 2005 to June 25, 2005.
# Received from: Dr. Anirban Mahanti.

# Topology: Dumbbel
# Default Settings are:
# Bottleneck Bandwidth: 45 Mbps
# Bottleneck2 Bandwidth: 45 Mbps
# Foreground flows: 1 Newreno and 1 Reno (RTT = 70 ms)
# Background flows (per bottleneck): 9 https and 1 ftp (RTT [20,460]) 
		# old settings (http RTT [20,220] ms and ftp RTT [20,460] ms)

# Bottleneck delay (oneway/for both bottleneck): 5 ms
# Midlink delay: 20ms
# Bottleneck queue type (both): DropTail 
# Bottleneck queue size (both): 250 packets
# Warmup period: 2 seconds
# Foreground flows start at warmup(2) + random(0-5)
# Background ftps start at: random(0-2)
# Background https think time: exponential(mean 500 ms) 

# ALL THE DEFAULT VALUES CAN BE CHANGED

set ns [new Simulator ]

# simulation parameters
set opt(simtime)     500.5;
set opt(warmup)      2.0;    	# foreground flows start warmup+random(0-5) sec.
set opt(bnbw)        45Mb  
set opt(bnbw2)       45Mb	  
set opt(bndelay)     0.005; 	# ONE WAY BOTTLENECK DELAY
set opt(bndelay2)    0.005; 	# ONE WAY BOTTLENECK DELAY
set opt(midbw)	   100Mb;	# Midlink BW
set opt(middelay)    0.020;	# Midlink delay
set opt(bw)          1000Mb;	# access link
set opt(pktsize)     1000;
set opt(queue)	   25000;	# Queue limit in packets
set opt(bnqueuetype) 0;		# 0 = DropTail, 1 = RED adaptive, 2 = RED static
set opt(bnqueuetype2) 0;		# 0 = DropTail, 1 = RED adaptive, 2 = RED static
set opt(bnqueuename) "DropTail";
set opt(bnqueuename2) "DropTail";
set opt(bnqueue)     250; 	# set appx. equal to bw*one way delay
set opt(bnqueue2)    250; 	# set appx. equal to bw*one way delay

set opt(fgrtt)       0.070;	# ONE WAY FOREGROUND DELAY for foregound TCP

set opt(losstype)	0;	# 0 = congestion loss, 1 = uniform, 2 = correlated
set opt(lossname) "CONGESTION";	
set opt(uniloss)	-0.10;	# negative implies no uniform wireless loss
set opt(p)		0.01;   # correlated loss parameter
set opt(q)		0.20;   # correlated loss parameter

# TCP parameters
set opt(tcptype)     TCP/Newreno;	#TCP type for foreground TCP and background TCP
set opt(tcptype2)    TCP/Reno;		#TCP type for foreground2 TCP
set opt(sinktype)    TCPSink
set opt(maxwin)      64; 		#Maximum window for background flows
set opt(fgmaxwin)    500;		#Maximum window for foreground & foreground2
set opt(httpbkgflows)   9;
set opt(ftpbkgflows)    1; 
set opt(httpbkgflows2)  -1;
set opt(ftpbkgflows2)   -1; 


set opt(tcpflows)	1; 	# number of foreground TCP flows
set opt(tcpflows2)	1; 	# number of foreground2 TCP flows


set opt(inactive)	0; 	# do not set it, program will set it
				# 0 = Both present
				# 1 = NewReno absent 
				# 2 = Reno absent


set opt(deterministic)	1;	# 1 = RTTs are chosen deterministically
				# 0 = RTTs are chosen randomly



# PROCESS INPUT ARGUMENTS
proc getopt {argc argv} {
    global opt

    for {set i 0} {$i < $argc} {incr i} {
        set arg [lindex $argv $i]
        if {[string range $arg 0 0] != "-"} continue

        set name [string range $arg 1 end]
        if {$name == "help"} {
            printopts
        } elseif {[info exists opt($name)]} {
            set val [lindex $argv [incr i]]
            global opt($name)
            set opt($name) $val
            puts "-$name set to $opt($name)"
        } else {
            puts "Invalid option: $name"
            printopts
        }
    }  
}
# Get the options and set simulation parameters accordingly
getopt [llength $argv] $argv

set opt(interval)    [expr $opt(simtime) * 0.75]; 	# collects data from the last 75% of the sim


# TCP specific details (Overriden by tweakTCP)
Agent/TCP set tcp_qs_recovery_ false
Agent/TCP set window_	   $opt(maxwin);
Agent/TCP set timestamps_	true
Agent/TCP set maxrto_	64
Agent/TCP set ecn_		0
Agent/TCP set packetSize_	960
Agent/TCP set tcpTick_	0.001

# Sets TCP parameters
proc tweakTCP { ctcp } {
	global opt
	$ctcp set tcp_qs_recovery_ false
	$ctcp set window_	$opt(maxwin);
	$ctcp set timestamps_	true
	$ctcp set windowInit_	1 # default is 2, but 1 accroding to our model
	$ctcp set windowOption_ 0 # default 1, 1 causes 1/cwnd increase.
	$ctcp set maxrto_	64
	$ctcp set ecn_		0; # MUST BE 0 to get pure TCP behavior
	$ctcp set packetSize_	960
	$ctcp set bugFix_	true
	$ctcp set tcpTick_	0.001
	$ctcp set singledup_ 0	# current default is 1. with 1, it sends packet for 1st and 2nd dupack
	$ctcp set newreno_changes_ 0 # usual behavior, otherwise send 1 packet on every 2 dupack
	$ctcp set newreno_changes1_ 0 # slow-but-steady, 1 makes impatient
	$ctcp set partial_window_deflation_ 1 # Default=1, 0 makes full window deflation
}

puts "";
# Queue specification for all link buffers 
Queue set limit_ [expr $opt(queue)]
Queue/DropTail	set   mean_pktsize_   1000
Queue/DropTail  set   queue_in_bytes_ false	

Queue/RED set bytes_ false
Queue/RED set queue_in_bytes_ false
Queue/RED set mean_pktsize_ 1000

if { $opt(bnqueuetype) == 2 } {
	set mnvalue	[expr $opt(bnqueue) / 3 ]
	set mxvalue	[expr 2 * $opt(bnqueue) / 3 ]

	Queue/RED set thresh_ [expr $mnvalue]
	Queue/RED set maxthresh_ [expr $mxvalue]
	Queue/RED set q_weight_	0.002
	Queue/RED set gentle_  true
	#Queue/RED set prob1_  0.003

	puts "RED minthresh: $mnvalue   maxthresh: $mxvalue";
}

set router0 [$ns node] 
set router1 [$ns node]


if { $opt(bnqueuetype) == 0 } { set opt(bnqueuename) "DropTail"; }
if { $opt(bnqueuetype) == 1 } { set opt(bnqueuename) "RED";  };		# RED adaptive/auto
if { $opt(bnqueuetype) == 2 } { set opt(bnqueuename) "RED"; };		# RED static

$ns duplex-link $router0 $router1 $opt(bnbw) $opt(bndelay) $opt(bnqueuename); 

# sets bnqueuetype for printing here and in data file
if { $opt(bnqueuetype) == 0 } { set opt(bnqueuename) "DropTail"; }
if { $opt(bnqueuetype) == 1 } { set opt(bnqueuename) "REDauto";  }
if { $opt(bnqueuetype) == 2 } { set opt(bnqueuename) "REDstatic"; }
puts "Bottleneck Queue: $opt(bnqueuename)";

# Bidirectional bottleneck queue with same config (for RED and DropTail)
$ns queue-limit $router0 $router1 $opt(bnqueue)
$ns queue-limit $router1 $router0 $opt(bnqueue)
puts "Bottleneck Queue Size: $opt(bnqueue)";


if { $opt(losstype) == 0 } { set opt(lossname) "CONGESTION"; }
if { $opt(losstype) == 1 } { set opt(lossname) "UNILOSS"; }
if { $opt(losstype) == 2 } { set opt(lossname) "COLOSS"; }
puts "Loss Type: $opt(lossname)"


# Adding Wireless Loss Module in Link $r0--$r1
proc add-unidrop {LossyLink} {
      global opt
    
      set em [new ErrorModel]
      $em unit pkt
      $em set  rate_ $opt(uniloss) 
      $em ranvar [new RandomVariable/Uniform]
      $em drop-target [new Agent/Null]
      $LossyLink errormodule $em
}

if { $opt(losstype) == 1 && $opt(uniloss) > 0.0 } {
        puts "Uniform Wireless Loss Rate: $opt(uniloss)";
	set flink [$ns link $router0 $router1]
        add-unidrop $flink
	# ADD REVERSE LINK ERROR LATER
} else {
        puts "No Wireless Loss";
}


### BOTTLENECK2 CONFIGURATION 

set router2 [$ns node] 
set router3 [$ns node]

if { $opt(bnqueuetype2) == 0 } { set opt(bnqueuename2) "DropTail"; }
if { $opt(bnqueuetype2) == 1 } { set opt(bnqueuename2) "RED";  };		# RED adaptive/auto
if { $opt(bnqueuetype2) == 2 } { set opt(bnqueuename2) "RED"; };		# RED static

$ns duplex-link $router2 $router3 $opt(bnbw2) $opt(bndelay2) $opt(bnqueuename2); 

# sets bnqueuetype for printing here and in data file
if { $opt(bnqueuetype2) == 0 } { set opt(bnqueuename2) "DropTail"; }
if { $opt(bnqueuetype2) == 1 } { set opt(bnqueuename2) "REDauto";  }
if { $opt(bnqueuetype2) == 2 } { set opt(bnqueuename2) "REDstatic"; }
puts "\nBottleneck2 Queue: $opt(bnqueuename2)";

# Bidirectional bottleneck queue with same config (for RED and DropTail)
$ns queue-limit $router2 $router3 $opt(bnqueue2)
$ns queue-limit $router3 $router2 $opt(bnqueue2)
puts "Bottleneck2 Queue Size: $opt(bnqueue2)";

#if { $opt(losstype) == 0 } { set opt(lossname) "CONGESTION"; }
#if { $opt(losstype) == 1 } { set opt(lossname) "UNILOSS"; }
#if { $opt(losstype) == 2 } { set opt(lossname) "COLOSS"; }
#puts "Loss Type: $opt(lossname)"
puts "Bottleneck2 default loss is CONGESTION"

### BOTTLENECK TWO end


# Connecting two bottlenecks
$ns duplex-link $router1 $router2 $opt(midbw) $opt(middelay) DropTail 



set rng [new RNG]
$rng seed 0

if { $opt(tcpflows) < 1 } {
	set opt(inactive) 1;  # NEWRENO inactive
	set opt(tcpflows) 1;
}

puts "";
# SETUP FOR FOREGROUND TCP FLOWS (TCP NEWRENO)
set flowid 0;
set fgcount 0;
for {set i 0} {$i < $opt(tcpflows)} {incr i 1} {
    set delay [expr $opt(fgrtt) / 2.0 ] ;	# RTT fixed
    puts "Foreground flow (NEWRENO) [expr $fgcount + 1] has RTT of $opt(fgrtt) ms"
    set linkdelay [expr $delay - $opt(bndelay) - $opt(bndelay2) -$opt(middelay) ]
    if {$linkdelay <= 0} {
        set linkdelay 1us;
        puts "Low link delay value";
    }

    set srctcp($i) [$ns node];
    $ns duplex-link $srctcp($i) $router0 $opt(bw) 1us DropTail 
    set snktcp($i) [$ns node]
    $ns duplex-link $snktcp($i) $router3 $opt(bw) $linkdelay DropTail
    set tcp($i) [$ns create-connection $opt(tcptype) $srctcp($i) \
            $opt(sinktype) $snktcp($i) $flowid]
    incr flowid 1;

    # sets TCP parameters
    tweakTCP $tcp($i)
    $tcp($i) set window_ $opt(fgmaxwin)
    $tcp($i) set maxcwnd_ [expr 2* $opt(fgmaxwin)]; # CAN RETHINK
    $tcp($i) set bugFix_  true;		# MUST BE TRUE FOR NEWRENO
    $tcp($i) set maxburst_ 4
    $tcp($i) set tcp_qs_recovery_ false
    $tcp($i) set qs_enabled_ false

    incr fgcount 1;
    puts "Foreground: $fgcount: $opt(tcptype)(fid=[expr $flowid - 1 ]) created ";	

    # attach FTP
    set ftp($i) [$tcp($i) attach-app FTP]
    if { $opt(inactive) == 1 } { 
	$ns at [expr $opt(simtime) ] "$ftp($i) start"; # starts at end, no effect
	puts "INACTIVE NEWRENO";
    } else {   
	$ns at [expr $opt(warmup) + [$rng uniform 0 5]] "$ftp($i) start"; # starts between [2,5]
    }
}



if { $opt(tcpflows2) < 1 } {
   set opt(inactive) 2;		# INACTIVE RENO
   set opt(tcpflows2) 1;
}


puts "";
# SETUP FOR FOREGROUND2 (RENO) TCP FLOWS 
set fgcount2 0;
for {set i $opt(tcpflows)} {$i < [expr $opt(tcpflows) + $opt(tcpflows2) ] } {incr i 1} {
    set delay [expr $opt(fgrtt) / 2.0 ] ;	# RTT fixed
    puts "Foreground flow (RENO) [expr $fgcount2 + 1] has RTT of $opt(fgrtt) ms"
    set linkdelay [expr $delay - $opt(bndelay) -$opt(bndelay2) - $opt(middelay) ]
    if {$linkdelay <= 0} {
        set linkdelay 1us;
        puts "Low link delay value\n";
    }

    set srctcp($i) [$ns node]
    $ns duplex-link $srctcp($i) $router0 $opt(bw) 1us DropTail 
    set snktcp($i) [$ns node]
    $ns duplex-link $snktcp($i) $router3 $opt(bw) $linkdelay DropTail
    set tcp($i) [$ns create-connection $opt(tcptype2) $srctcp($i) \
            $opt(sinktype) $snktcp($i) $flowid]
    incr flowid 1;

    # sets TCP parameters 	
    tweakTCP $tcp($i)
    $tcp($i) set window_ $opt(fgmaxwin)
    $tcp($i) set window_ $opt(fgmaxwin)
    $tcp($i) set maxcwnd_ [expr 2* $opt(fgmaxwin)]
    $tcp($i) set bugFix_  false;	# MUST BE FALSE FOR RENO
    $tcp($i) set maxburst_ 0;		
    $tcp($i) set tcp_qs_recovery_ false
    $tcp($i) set qs_enabled_ false

    incr fgcount2 1;
    puts "Foreground2: $fgcount2: $opt(tcptype2)(fid=[expr $flowid - 1]) created ";	


    # attach FTP
    set ftp($i) [$tcp($i) attach-app FTP]
    if { $opt(inactive) == 2  } {
    	$ns at [expr $opt(simtime) ] "$ftp($i) start"; # starts at end, no effect
	puts "INACTIVE RENO";
    } else {
    	$ns at [expr $opt(warmup) + [$rng uniform 0 5]] "$ftp($i) start"; # starts between [2,5]
    }
}

# SETUP FOR BACKGROUND FTP FLOWS (uses NewReno as TCP)
puts "";
set rngd1 [new RNG]
$rngd1 seed 0
set stfrom 	[expr $opt(tcpflows) + $opt(tcpflows2) ];
set endto	[expr $opt(tcpflows) + $opt(tcpflows2) + $opt(ftpbkgflows) ];
set ftpcount 0;

# background flows is assumed to be more than 1
set rttinc 100
set rttinc [expr 440 / [expr $opt(ftpbkgflows) + 1] ]; # (460 -20)/(flowno+1)

for {set i $stfrom } {$i < $endto} {incr i 1} {
	incr ftpcount 1;
	set delay [expr 20 + $rttinc * $ftpcount ] ; # sets RTT deterministically
	if { $opt(deterministic) == 0 } {
		set delay [expr [$rngd1 integer 441] + 20]; 	# RTT chosen randomly from [20, 460]
	}
      	puts "Background FTP $ftpcount (fid=$flowid) has a rtt of $delay ms"
	set delay [expr $delay / 2.0 ];	# making one way delay
      	set linkdelay [expr ($delay - $opt(bndelay)) / 2.0 ]; # see the topology next
	if {!$linkdelay} {
        	set linkdelay 1us
    	} else {
        	append linkdelay ms
    	}

	set srctcp($i) [$ns node]
	$ns duplex-link $srctcp($i) $router0 $opt(bw) $linkdelay DropTail 
    	set snktcp($i) [$ns node]
    	$ns duplex-link $snktcp($i) $router1 $opt(bw) $linkdelay DropTail
    	set tcp($i) [$ns create-connection $opt(tcptype) $srctcp($i) \
	    $opt(sinktype) $snktcp($i) $flowid]
    	incr flowid 1;

    	# setup TCP parameters
    	tweakTCP $tcp($i)
    	$tcp($i) set maxcwnd_ [expr 2* $opt(maxwin)]  # DIFFERENT FROM FOREGROUND NEWRENO
    	$tcp($i) set bugFix_  true;		# MUST BE TRUE FOR NEWRENO
    	$tcp($i) set maxburst_ 4
    	$tcp($i) set tcp_qs_recovery_ false
    	$tcp($i) set qs_enabled_ false

	# attach FTP
	set ftp($i) [$tcp($i) attach-app FTP]
	$ns at [$rng uniform 0 2] "$ftp($i) start";  # starts between [0,2]
}


# BOTTLENECK2: SETUP FOR BACKGROUND FTP FLOWS (uses NewReno as TCP)
puts "";
if { $opt(ftpbkgflows2) < 0 } {
   set opt(ftpbkgflows2) $opt(ftpbkgflows);		# INACTIVE RENO
}
set stfrom 	[expr $opt(tcpflows) + $opt(tcpflows2) + $opt(ftpbkgflows) ];
set endto	[expr $opt(tcpflows) + $opt(tcpflows2) + $opt(ftpbkgflows) + $opt(ftpbkgflows2) ];
set ftpcount2 0;

# background flows is assumed to be more than 1
set rttinc 100
set rttinc [expr 440 / [expr $opt(ftpbkgflows2) + 1] ]; # (460 -20)/(flowno+1)

for {set i $stfrom } {$i < $endto} {incr i 1} {
	incr ftpcount2 1;
	set delay [expr 20 + $rttinc * $ftpcount2 ] ; # sets RTT deterministically
	if { $opt(deterministic) == 0 } {
		set delay [expr [$rngd1 integer 441] + 20]; 	# RTT chosen randomly from [20, 460]
	}
      puts "Background2 FTP $ftpcount2 (fid=$flowid) has a rtt of $delay ms"
	set delay [expr $delay / 2.0 ];	# making one way delay
      	set linkdelay [expr ($delay - $opt(bndelay2)) / 2.0 ]; # see the topology next
	if {!$linkdelay} {
        	set linkdelay 1us
    	} else {
        	append linkdelay ms
    	}

	set srctcp($i) [$ns node]
	$ns duplex-link $srctcp($i) $router2 $opt(bw) $linkdelay DropTail 
    	set snktcp($i) [$ns node]
    	$ns duplex-link $snktcp($i) $router3 $opt(bw) $linkdelay DropTail
    	set tcp($i) [$ns create-connection $opt(tcptype) $srctcp($i) \
	    $opt(sinktype) $snktcp($i) $flowid]
    	incr flowid 1;

    	# setup TCP parameters
    	tweakTCP $tcp($i)
    	$tcp($i) set maxcwnd_ [expr 2* $opt(maxwin)]  # DIFFERENT FROM FOREGROUND NEWRENO
    	$tcp($i) set bugFix_  true;		# MUST BE TRUE FOR NEWRENO
    	$tcp($i) set maxburst_ 4
    	$tcp($i) set tcp_qs_recovery_ false
    	$tcp($i) set qs_enabled_ false

	# attach FTP
	set ftp($i) [$tcp($i) attach-app FTP]
	$ns at [$rng uniform 0 2] "$ftp($i) start";  # starts between [0,2]
}
### Background2 FTP end



# HTTP TRAFFIC GENERATION

# Exponentially distributed inter-request time ("think-time") with mean 0.5 seconds
set rnginterval [new RNG]
$rnginterval seed 0
set interval [new RandomVariable/Exponential]
$interval use-rng $rnginterval
$interval set avg_ 0.500  # previous default was 0.5

# Web page distribution is Pareto with mean 48 KB, shape 1.2
set rngsize [new RNG]
$rngsize seed 0
set size [new RandomVariable/Pareto]
#set size [new RandomVariable/Exponential]
$size use-rng $rngsize
$size set avg_ [expr 48*960]
$size set shape_ 1.2; # generate mice traffic 

set rttinc 100
set rttinc [expr 440 / [expr $opt(httpbkgflows) + 1] ]; # (460 - 20)/(flowno+1)
set httpcount 0;

set flowid 2000;

puts "";
for {set i 0} {$i < $opt(httpbkgflows)} {incr i 1} {
    incr httpcount 1;
    # set up tcp agent for http server
    set http_server_tcp($i) [new Agent/TCP/FullTcp]
    tweakTCP $http_server_tcp($i)
    $http_server_tcp($i) set window_ $opt(maxwin)
    $http_server_tcp($i) set packetSize_ $opt(pktsize)
    $http_server_tcp($i) set segsize_ 960
    incr flowid 1;
    $http_server_tcp($i) set fid_ [expr $flowid]
    
    # setup nodes, links, agents, and connect

    set delay [expr 20 + $rttinc * $httpcount ];   # sets RTT deterministically
    if { $opt(deterministic) == 0 } {
	set delay [expr [$rngd1 integer 441] + 20]; 	# RTT chosen randomly from [20, 220]
    }

    puts "Bkg. HTTP flow $httpcount (fid=[expr $flowid]) has RTT of $delay ms"
    set delay [expr $delay / 2.0 ];			# making one way delay
    set linkdelay [expr ($delay - $opt(bndelay))]
    if {!$linkdelay} {
        set linkdelay 1us
    } else {
        append linkdelay ms
    }

    set websrc($i) [$ns node]    
    $ns duplex-link $websrc($i) $router0 $opt(bw) $linkdelay DropTail

    # set up tcp agent for http client
    set http_client_tcp($i) [new Agent/TCP/FullTcp]
    tweakTCP $http_client_tcp($i)
    $http_client_tcp($i) set window_ $opt(maxwin)
    $http_client_tcp($i) set packetSize_ $opt(pktsize)
    $http_client_tcp($i) set segsize_ 960
    $http_client_tcp($i) set fid_ [expr $flowid]

    set websnk($i) [$ns node]
    $ns duplex-link $websnk($i) $router1 $opt(bw) 1us DropTail 

    
    # attach server tcp agent on a node at the left hand side of the dumb-bell
    $ns attach-agent $websrc($i) $http_server_tcp($i)

    # attach client tcp agent on a node at the right hand side 
    $ns attach-agent $websnk($i) $http_client_tcp($i)

    # connect the client to the server
    set http_agent($i) [$ns connect $http_client_tcp($i) $http_server_tcp($i)]
    # set the server at passive mode (waiting forrequests from the client)
    $http_server_tcp($i) listen
    #$http_client_tcp($i) listen
  
    # Create application layer modules to sit on 
    # top of the tcp agents and specify the connnection between them
    set http_client($i) [new Application/TcpApp $http_client_tcp($i)]
    set http_server($i) [new Application/TcpApp $http_server_tcp($i)]
    $http_client($i) connect $http_server($i)
  
  
    set time($i) [$interval value]
    $ns at $time($i) "$http_client($i) send 960 \" $http_server($i) request-handler $i \" "
    #puts "*** first request for http flow $i scheduled at $time($i)"
}



### BACKGROUND2: HTTP TRAFFIC 

if { $opt(httpbkgflows2) < 0 } {
   set opt(httpbkgflows2) $opt(httpbkgflows);		# INACTIVE RENO
}

set stfrom 	[expr $opt(httpbkgflows) ];
set endto	[expr $opt(httpbkgflows) + $opt(httpbkgflows2) ];

set rttinc 100
set rttinc [expr 440 / [expr $opt(httpbkgflows2) + 1] ]; # (460 - 20)/(flowno+1)
set httpcount2 0;

puts "";
for {set i $stfrom} {$i < $endto} {incr i 1} {
    incr httpcount2 1;
    # set up tcp agent for http server
    set http_server_tcp($i) [new Agent/TCP/FullTcp]
    tweakTCP $http_server_tcp($i)
    $http_server_tcp($i) set window_ $opt(maxwin)
    $http_server_tcp($i) set packetSize_ $opt(pktsize)
    $http_server_tcp($i) set segsize_ 960
    incr flowid 1;
    $http_server_tcp($i) set fid_ [expr $flowid]
    
    # setup nodes, links, agents, and connect

    set delay [expr 20 + $rttinc * $httpcount2 ];   # sets RTT deterministically
    if { $opt(deterministic) == 0 } {
	set delay [expr [$rngd1 integer 441] + 20]; 	# RTT chosen randomly from [20, 220]
    }

    puts "Bkg.2 HTTP flow $httpcount2 (fid=[expr $flowid]) has RTT of $delay ms"
    set delay [expr $delay / 2.0 ];			# making one way delay
    set linkdelay [expr ($delay - $opt(bndelay2))]
    if {!$linkdelay} {
        set linkdelay 1us
    } else {
        append linkdelay ms
    }

    set websrc($i) [$ns node]    
    $ns duplex-link $websrc($i) $router2 $opt(bw) $linkdelay DropTail

    # set up tcp agent for http client
    set http_client_tcp($i) [new Agent/TCP/FullTcp]
    tweakTCP $http_client_tcp($i)
    $http_client_tcp($i) set window_ $opt(maxwin)
    $http_client_tcp($i) set packetSize_ $opt(pktsize)
    $http_client_tcp($i) set segsize_ 960
    $http_client_tcp($i) set fid_ [expr $flowid]

    set websnk($i) [$ns node]
    $ns duplex-link $websnk($i) $router3 $opt(bw) 1us DropTail 

    
    # attach server tcp agent on a node at the left hand side of the dumb-bell
    $ns attach-agent $websrc($i) $http_server_tcp($i)

    # attach client tcp agent on a node at the right hand side 
    $ns attach-agent $websnk($i) $http_client_tcp($i)

    # connect the client to the server
    set http_agent($i) [$ns connect $http_client_tcp($i) $http_server_tcp($i)]
    # set the server at passive mode (waiting forrequests from the client)
    $http_server_tcp($i) listen
    #$http_client_tcp($i) listen
  
    # Create application layer modules to sit on 
    # top of the tcp agents and specify the connnection between them
    set http_client($i) [new Application/TcpApp $http_client_tcp($i)]
    set http_server($i) [new Application/TcpApp $http_server_tcp($i)]
    $http_client($i) connect $http_server($i)
  
  
    set time($i) [$interval value]
    $ns at $time($i) "$http_client($i) send 960 \" $http_server($i) request-handler $i \" "
    #puts "*** first request for http flow $i scheduled at $time($i)"
}

### Background2: HTTP traffic end

    
Application/TcpApp instproc request-handler { index } {
    global ns http_server http_client size

    set page_size [expr (1*[$size value])]
    #set page_size [expr (1000*[$size value])/1000]
    #puts "http_server($index) receives a request at time [$ns now] for $page_size bytes"
    $http_server($index) send $page_size "$http_client($index) request-generator $index"
    #puts "@@@ Answer for http flow $index scheduled now"
}

Application/TcpApp instproc request-generator { index } {
    global ns http_server http_client interval

    #puts "http_client($index) sends a new request at time [$ns now] "
    set time [expr [$ns now] + [$interval value]]
    $ns at $time "$http_client($index) send 960 \" $http_server($index) request-handler $index \" "
    #puts "*** Request for http flow $index scheduled at $time"
    #$http_client($index) send 960 "$http_server($index) request-handler $index"
}

# HTTP CONNECTIONS CREATED


set bottleneck [$ns link $router0 $router1]; 		#link to collect stats on
set qmon [$ns monitor-queue $router0 $router1 ""]
set fmon [$ns makeflowmon Fid]
$ns attach-fmon $bottleneck $fmon

set bottleneck2 [$ns link $router2 $router3]; 		#link to collect stats on
set qmon2 [$ns monitor-queue $router2 $router3 ""]
set fmon2 [$ns makeflowmon Fid]
$ns attach-fmon $bottleneck2 $fmon2




$ns at [expr $opt(simtime) - $opt(interval) - 0.5] "resetLinkDump"
$ns at [expr $opt(simtime) - 0.5]  "linkDump"
$ns at $opt(simtime) "finish"
   
proc finish {} {
    global ns
    $ns flush-trace
    puts "Simulation Done"
    exit 0
}



# We monitor two foreground flows: one foreground NewReno and one foreground Reno
set nrf	[expr 0]
set rf	[expr $opt(tcpflows)]

# resets loss event trace variables
$tcp($nrf)	set 	letracestart [expr $opt(simtime) - $opt(interval) - 0.5 ];
$tcp($rf)	set 	letracestart [expr $opt(simtime) - $opt(interval) - 0.5 ];



# RTT MONITORING

# traces RTT on every 125ms for monitored NewReno flow
$ns at [expr $opt(simtime) - $opt(interval) - 0.5 ] "nrreportRTT  $tcp($nrf)"

set nrsumRTT		[expr  0.0000001]
set nrcountRTT		[expr  0]
set nravgRTT		[expr  0]
set nrTickValue		[$tcp($nrf) set tcpTick_]

# Monitors RTT for the foreground flow no 0
proc nrreportRTT {ctcp} {
    global ns nrsumRTT nrcountRTT nravgRTT nrTickValue

    set temprtt [$ctcp set rtt_]
    set curtime [$ns now]

    set nrsumRTT [expr $nrsumRTT + $temprtt * $nrTickValue ]
    set nrcountRTT [expr $nrcountRTT + 1]
    set nravgRTT [expr $nrsumRTT/$nrcountRTT]
    # puts "$nrcountRTT:$curtime \t $nravgRTT \t $temprtt"

    $ns at [expr $curtime + .125] "nrreportRTT  $ctcp"
}


# traces RTT on every 125ms for monitored Reno flow
$ns at [expr $opt(simtime) - $opt(interval) - 0.5 ] "rreportRTT  $tcp($rf)"

set rsumRTT		[expr  0.0000001]
set rcountRTT		[expr  0]
set ravgRTT		[expr  0]
set rTickValue		[$tcp($rf) set tcpTick_]

# Monitors RTT for the foreground flow no 0
proc rreportRTT {ctcp} {
    global ns rsumRTT rcountRTT ravgRTT rTickValue

    set temprtt [$ctcp set rtt_]
    set curtime [$ns now]

    set rsumRTT [expr $rsumRTT + $temprtt * $rTickValue ]
    set rcountRTT [expr $rcountRTT + 1]
    set ravgRTT [expr $rsumRTT/$rcountRTT]
    # puts "$rcountRTT:$curtime \t $ravgRTT \t $rtemprtt"

    $ns at [expr $curtime + .125] "rreportRTT  $ctcp"
}




set stat_file [open bnlink.data w]


# RESETS MONITORED LINK STATS
proc resetLinkDump { } {
    global ns qmon fmon qmon2 fmon2
    foreach f [$fmon flows] {
        $f reset;
    }
    $qmon reset;
    $fmon reset;

    foreach f2 [$fmon2 flows] {
        $f2 reset;
    }
    $qmon2 reset;
    $fmon2 reset;
}


# LINK DUMP PROCEDURE
puts "";
proc linkDump {} {

    global ns util opt stat_file bottleneck fmon2 qmon2 fmon qmon nravgRTT ravgRTT nrf rf tcp
    
    # overall link statistics
    set pdep [$qmon set pdepartures_]; 
    set parrv [$qmon set parrivals_]
    set pdrops [$qmon set pdrops_]
    set bdep [$qmon set bdepartures_];

    set pdrops2 [$qmon2 set pdrops_]
    set pdropstotal [expr $pdrops2 + $pdrops ]

    set drprt [expr 0]; 	# not calculated
    if {$parrv > 0} {
	set drprt [expr [expr $pdropstotal*1.0]/$parrv]; 	# this is not in percentage
    }

    # get link capacity
    set capac [[$bottleneck link] set bandwidth_];
    set util [expr 0.0008*$bdep/[expr 0.0001*$opt(interval)*$capac]]
    

    # monitored flow statistics	
    for {set id 0} {$id < [expr $opt(tcpflows) + $opt(tcpflows2)]} {incr id 1} {
	set frate($id) 0;
	set fdrprt($id) 0;
	set farrv($id) 0;
	set fdep($id) 0;
	set fbdep($id) 0;
	set fpdrops($id) 0;
    }

    for {set id 0} {$id < [expr $opt(tcpflows) + $opt(tcpflows2)]} {incr id 1} {
	set fpdrops2($id) 0;
    }


	
    # get per flow stats for the competing flows
    foreach f [$fmon flows] {
	set id [$f set flowid_];
	set farrv($id) [$f set parrivals_];
	set fdep($id) [$f set pdepartures_];
	set fbdep($id) [$f set bdepartures_];
	set fpdrops($id) [$f set pdrops_];

	$f reset;
    }


    foreach f2 [$fmon2 flows] {
	set id [$f2 set flowid_];
	set fpdrops2($id) [$f2 set pdrops_];

	$f2 reset;
    }




    # generate output
    set statsprint [$ns now];

    # calculating monitored flows packet drop rate, not percentage
    set fpdropstotal($nrf) [expr $fpdrops2($nrf) + $fpdrops($nrf) ]
    set fdrprt($nrf) 0
    if {$farrv($nrf) > 0} {
	set fdrprt($nrf) [expr [expr $fpdropstotal($nrf) * 1.0 ]/$farrv($nrf)];
    }
    set fdrprt($rf) 0
    set fpdropstotal($rf) [expr $fpdrops2($rf) + $fpdrops($rf) ]

    if {$farrv($rf) > 0} {
	set fdrprt($rf) [expr [expr $fpdropstotal($rf) * 1.0 ]/$farrv($rf)];
    }

    set frate($nrf) [expr $fdep($nrf) * 1.0 / $opt(interval)];		# pkts/sec
    set frate($nrf) [expr $frate($nrf) * $nravgRTT ];			# pkts/RTT
	
    set frate($rf) [expr $fdep($rf) * 1.0 / $opt(interval)];		# pkts/sec
    set frate($rf) [expr $frate($rf) * $ravgRTT ];			# pkts/RTT
	    

    set nrto [$tcp($nrf) set countto]
    set nrifr [$tcp($nrf) set countifr]
    set nrfr [$tcp($nrf) set countfr]
    set nre2 [expr $nrto+$nrifr+$nrfr]
    set nracked [$tcp($nrf) set countpktacked]
    set nre [$tcp($nrf) set countlossevent]
    set nrl [$tcp($nrf) set counttotalloss]
    set nrsnt [expr $nracked + $nrl]
    puts "NewReno Loss Events: $nre  Loss: $nrl  LinkLoss: $fpdropstotal($nrf) TO: $nrto"
    puts "NRsent: $nrsnt NRacked: $nracked LinkDepart: $fdep($nrf) avgRTT: $nravgRTT"

    set lert($nrf) 0
    set lrt($nrf) 0
    set tp($nrf) 0
    if { $nrsnt > 0 } {
    	set lert($nrf) [expr 1.0 * $nre / $nrsnt ]
	set lert2($nrf) [expr 1.0 * $nre2 / $nrsnt ]
    	set lrt($nrf) [expr 1.0 * $nrl / $nrsnt ]
	set tp($nrf) [expr 1.0 * $nracked * $nravgRTT / $opt(interval) ]
	# More throughput in different format
	set nrtppr [expr 1.0 * $nracked * $nravgRTT / $opt(interval) ];	# Packets/RTT
        set nrtpps [expr 1.0 * $nracked / $opt(interval) ];		# Packets/Sec.
        set nrtpkbps [expr 8.0 * $nracked  / $opt(interval) ];		# Kbps
        set nrtpmbps [expr 8.0 * $nracked  / $opt(interval) / 1000.0 ];	# Mbps
    }



    set rto [$tcp($rf) set countto]
    set racked [$tcp($rf) set countpktacked]
    set re [$tcp($rf) set countlossevent]
    set rl [$tcp($rf) set counttotalloss]
    set rsnt [expr $racked + $rl]
    puts "Reno Loss Events: $re  Loss: $rl LinkLoss: $fpdropstotal($rf) TO: $rto"
    puts "Rsnt: $rsnt Racked: $racked  LinkDepart: $fdep($rf) avgRTT: $ravgRTT\n"

    set lert($rf) 0
    set lrt($rf) 0
    set tp($rf) 0
    if { $rsnt > 0 } {
    	set lert($rf) [expr 1.0 * $re / $rsnt ]
    	set lrt($rf) [expr 1.0 * $rl / $rsnt ]
	set tp($rf) [expr 1.0 * $racked * $ravgRTT / $opt(interval) ]
	# More throughput in different format
	set rtppr [expr 1.0 * $racked * $ravgRTT / $opt(interval) ];	# Packets/RTT
        set rtpps [expr 1.0 * $racked / $opt(interval) ];		# Packets/Sec.
        set rtpkbps [expr 8.0 * $racked  / $opt(interval) ];		# Kbps
        set rtpmbps [expr 8.0 * $racked  / $opt(interval) / 1000.0 ];	# Mbps
    }


    set statsprint [format "NRer2 %9.7f NRer %9.7f NRlr %7.5f NRlrL %7.5f NRtp %7.3f NRtpL %7.3f NRrtt %6.4f" $lert2($nrf) $lert($nrf) $lrt($nrf) $fdrprt($nrf) $tp($nrf) $frate($nrf) $nravgRTT];
    set statsprint [format "%s NRe %d NRl %d NRlL %d" $statsprint $nre $nrl $fpdrops($nrf)];
    set statsprint [format "%s NRacked %d NRsnt %d NRdepL %d NRarvL %d NRto %d" $statsprint $nracked $nrsnt $fdep($nrf) $farrv($nrf) $nrto];
    set statsprint [format "%s NRtppr %7.3f NRtpps %7.2f NRtpkbps %7.2f NRtpmbps %7.4f" $statsprint $nrtppr $nrtpps $nrtpkbps $nrtpmbps];
    set statsprint [format "%s NRfr %d NRifr %d NRe2 %d" $statsprint $nrfr $nrifr $nre2];


    set statsprint [format "%s Rer %9.7f Rlr %7.5f RlrL %7.5f Rtp %7.3f RtpL %7.3f Rrtt %6.4f" $statsprint $lert($rf) $lrt($rf) $fdrprt($rf) $tp($rf) $frate($rf) $ravgRTT]
    set statsprint [format "%s Re %d Rl %d RlL %d" $statsprint $re $rl $fpdrops($rf)];
    set statsprint [format "%s Racked %d Rsnt %d RdepL %d RarvL %d Rto %d" $statsprint $racked $rsnt $fdep($rf) $farrv($rf) $rto];
    set statsprint [format "%s Rtppr %7.3f Rtpps %7.2f Rtpkbps %7.2f Rtpmbps %7.4f" $statsprint $rtppr $rtpps $rtpkbps $rtpmbps];


    set statsprint [format "%s Lnkdrt %7.5f U %0.4f WatchTime %6.2f " $statsprint $drprt $util $opt(interval) ]    

    set flowcount1 [expr $opt(ftpbkgflows) + $opt(httpbkgflows) ];
    set flowcount2 [expr $opt(ftpbkgflows2) + $opt(httpbkgflows2) ];
    set flowcount [expr $opt(ftpbkgflows) + $opt(httpbkgflows) + $opt(ftpbkgflows2) + $opt(httpbkgflows2) ];
    set statsprint [format "%s bnqueuename %s bnbw %s bnqueue %d flow1 %d ftp %d http %d fgRTT %6.4f" $statsprint $opt(bnqueuename)  $opt(bnbw) $opt(bnqueue) $flowcount1 $opt(ftpbkgflows) $opt(httpbkgflows) $opt(fgrtt) ];
    set statsprint [format "%s bnqueuename2 %s bnbw2 %s bnqueue2 %d flow2 %d ftp2 %d http2 %d " $statsprint $opt(bnqueuename2)  $opt(bnbw2) $opt(bnqueue2) $flowcount2 $opt(ftpbkgflows2) $opt(httpbkgflows2) ];
    set statsprint [format "%s flow %d losstype %s uniloss %5.4f p %5.3f q %5.3f" $statsprint $flowcount $opt(lossname) $opt(uniloss) $opt(p) $opt(q)];



    puts "Printing Simulation Results:"
    puts $statsprint;	
    puts $stat_file $statsprint;
    
    $fmon reset;
    $qmon reset;
    $fmon2 reset;
    $qmon2 reset;

    $ns at [expr [$ns now] + $opt(interval)] "linkDump"
}

$ns run
